import os

from biolib.app.utils import run_job
from biolib.biolib_api_client import JobState
from biolib.biolib_api_client.app_types import App, AppVersion
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.app.app_result import AppResult
from biolib.biolib_api_client.biolib_app_api import BiolibAppApi
from biolib.biolib_binary_format import ModuleInput
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger


class BioLibApp:

    def __init__(self, uri: str):
        app_response = BiolibAppApi.get_by_uri(uri)
        self._app: App = app_response['app']
        self._app_uri = app_response['app_uri']
        self._app_version: AppVersion = app_response['app_version']

        logger.info(f'Loaded project {self._app_uri}')

    def __str__(self) -> str:
        return self._app_uri

    @property
    def uuid(self) -> str:
        return self._app['public_id']

    def cli(self, args=None, stdin=None, files=None):
        module_input_serialized = self._get_serialized_module_input(args, stdin, files)

        job = BiolibJobApi.create(self._app_version['public_id'])
        BiolibJobApi.update_state(job['public_id'], JobState.IN_PROGRESS.value)

        try:
            module_output = run_job(job, module_input_serialized)
            try:
                BiolibJobApi.update_state(job_id=job['public_id'], state=JobState.COMPLETED.value)
            except Exception as error:  # pylint: disable=broad-except
                logger.warning(f'Could not update job state to completed:\n{error}')

            return AppResult(
                exitcode=module_output['exit_code'],
                stderr=module_output['stderr'],
                stdout=module_output['stdout'],
                files=module_output['files']
            )

        except BioLibError as exception:
            logger.error(f'Compute failed with: {exception.message}')
            try:
                BiolibJobApi.update_state(job_id=job['public_id'], state=JobState.FAILED.value)
            except Exception as error:  # pylint: disable=broad-except
                logger.warning(f'Could not update job state to failed:\n{error}')

            raise exception

    def __call__(self, *args, **kwargs):
        if not args and not kwargs:
            self.cli()

        else:
            raise BioLibError('''
Calling an app directly with app() is currently being reworked.
To use the previous functionality, please call app.cli() instead. 
Example: "app.cli('--help')"
''')

    @staticmethod
    def _get_serialized_module_input(args=None, stdin=None, files=None) -> bytes:
        if args is None:
            args = []

        if stdin is None:
            stdin = b''

        if isinstance(args, str):
            args = list(filter(lambda p: p != '', args.split(' ')))

        if not isinstance(args, list):
            raise Exception('The given input arguments must be list or str')

        if isinstance(stdin, str):
            stdin = stdin.encode('utf-8')

        if files is None:
            files = []
            for idx, arg in enumerate(args):
                if os.path.isfile(arg):
                    files.append(arg)
                    args[idx] = arg.split('/')[-1]

        cwd = os.getcwd()
        files_dict = {}

        for file in files:
            path = file
            if not file.startswith('/'):
                # make path absolute
                path = cwd + '/' + file

            arg_split = path.split('/')
            file = open(path, 'rb')
            path = '/' + arg_split[-1]

            files_dict[path] = file.read()
            file.close()

        module_input_serialized: bytes = ModuleInput().serialize(stdin=stdin, arguments=args, files=files_dict)
        return module_input_serialized



import json
import os
import shutil
import time
import uuid

import appdirs  # type: ignore

from biolib.biolib_errors import BioLibError
from biolib.compute_node.job_worker.cache_types import CacheStateDict, UuidStr, StoragePartition
from biolib.typing_utils import Dict, List, Optional


class CacheStateError(BioLibError):
    pass


class CacheState:
    _cache_dir: str = appdirs.user_cache_dir(appname='pybiolib', appauthor='biolib')
    _state_path: str = f'{_cache_dir}/cache-state.json'
    _state_lock_path: str = f'{_state_path}.lock'

    def __init__(self):
        self._state: Optional[CacheStateDict] = None

    def __enter__(self) -> CacheStateDict:
        self._acquire_state_lock()
        if os.path.exists(self._state_path):
            with open(self._state_path, mode='r') as file:
                self._state = json.loads(file.read())
        else:
            self._state = CacheStateDict(
                large_file_systems={},
                storage_partitions=self._get_storage_partitions_from_env(),
            )
            with open(CacheState._state_path, mode='w') as file:
                file.write(json.dumps(self._state))

        # Check for type checking
        if self._state is None:
            raise CacheStateError('Internal state is not defined')

        return self._state

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        with open(CacheState._state_path, mode='w') as file:
            file.write(json.dumps(self._state))

        self._release_state_lock()

    @staticmethod
    def get_tmp_storage_paths() -> List[str]:
        lfs_tmp_storage_path_env_key = 'BIOLIB_LFS_TMP_STORAGE_PATHS'
        lfs_tmp_storage_paths = os.environ.get(lfs_tmp_storage_path_env_key)
        if lfs_tmp_storage_paths is None:
            raise CacheStateError(f'Environment variable "{lfs_tmp_storage_path_env_key}" not set')

        return lfs_tmp_storage_paths.split(',')

    @staticmethod
    def _get_storage_partitions_from_env() -> Dict[UuidStr, StoragePartition]:
        lfs_storage_path_env_key = 'BIOLIB_LFS_STORAGE_PATHS'
        lfs_storage_paths = os.environ.get(lfs_storage_path_env_key)
        if lfs_storage_paths is None:
            raise CacheStateError(f'Environment variable "{lfs_storage_path_env_key}" not set')

        storage_states: Dict[UuidStr, StoragePartition] = {}
        for lfs_storage_path in lfs_storage_paths.split(','):
            if not os.path.isdir(lfs_storage_path):
                raise CacheStateError(f'LFS storage path {lfs_storage_path} is not a directory')

            uuid_str = str(uuid.uuid4())
            disk_usage = shutil.disk_usage(lfs_storage_path)
            storage_states[uuid_str] = StoragePartition(
                allocated_size_bytes=0,
                path=lfs_storage_path,
                total_size_bytes=disk_usage.free,
                uuid=uuid_str,
            )

        return storage_states

    @staticmethod
    def _acquire_state_lock() -> None:
        timeout_seconds = 5.0
        seconds_to_sleep = 0.5
        while os.path.exists(CacheState._state_lock_path):
            time.sleep(seconds_to_sleep)
            timeout_seconds -= seconds_to_sleep
            if timeout_seconds < 0:
                raise CacheStateError('Cache state timed out waiting for lock file')

        os.makedirs(CacheState._cache_dir, exist_ok=True)
        lock_file = open(CacheState._state_lock_path, mode='x')
        lock_file.close()

    @staticmethod
    def _release_state_lock() -> None:
        if os.path.exists(CacheState._state_lock_path):
            os.remove(CacheState._state_lock_path)
        else:
            raise CacheStateError('Cache state was not locked.')



from urllib.parse import urlparse
from biolib.biolib_api_client import RemoteHost
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger
from biolib.compute_node.remote_host_proxy import RemoteHostProxy
from biolib.compute_node.webserver.webserver_types import WebserverConfig
from biolib.typing_utils import List



def start_enclave_remote_hosts(config: WebserverConfig) -> None:
    logger.debug('Starting Docker network for enclave remote host proxies')
    docker = BiolibDockerClient.get_docker_client()
    public_network = docker.networks.create(
        driver='bridge',
        internal=False,
        name='biolib-enclave-remote-hosts-network',
    )

    biolib_remote_hosts: List[RemoteHost] = []
    biolib_remote_host_proxies: List[RemoteHostProxy] = []
    base_hostname = urlparse(config['base_url']).hostname
    # Make sure base_hostname is not None for typing reasons
    if not base_hostname:
        raise Exception('Base hostname not set, likely due to base url not being set. This is required in enclaves')

    if base_hostname == 'biolib.com':
        biolib_remote_hosts.append(RemoteHost(hostname='containers.biolib.com'))
    elif base_hostname == 'staging-elb.biolib.com':
        biolib_remote_hosts.append(RemoteHost(hostname='containers.staging.biolib.com'))

    # Allow reaching backend for self registering and job creation
    biolib_remote_hosts.append(RemoteHost(hostname=base_hostname))
    # For downloading container image layers from S3 URLs returned by ECR proxy
    biolib_remote_hosts.append(RemoteHost(
        hostname=f"prod-{config['ecr_region_name']}-starport-layer-bucket.s3.{config['ecr_region_name']}.amazonaws.com"
    ))
    # For downloading source files zip
    biolib_remote_hosts.append(RemoteHost(
        hostname=f"{config['s3_general_storage_bucket_name']}.s3.amazonaws.com"
    ))

    logger.debug('Starting enclave remote host proxies')
    for remote_host in biolib_remote_hosts:
        remote_host_proxy = RemoteHostProxy(remote_host, public_network, internal_network=None, job_id=None)
        remote_host_proxy.start()
        biolib_remote_host_proxies.append(remote_host_proxy)

    logger.debug('Writing to enclave /etc/hosts')
    with open('/etc/hosts', mode='a') as hosts_file:
        for proxy in biolib_remote_host_proxies:
            ip_address = proxy.get_ip_address_on_network(public_network)
            hosts_file.write(f'\n{ip_address} {proxy.hostname}')



import argparse
import logging
import os
import sys
import platform

from docker.errors import ImageNotFound  # type: ignore

from biolib import cli_utils, utils, biolib_errors
from biolib.app import BioLibApp
from biolib.app.utils import run_job
from biolib.biolib_push import push_application
from biolib.biolib_api_client import Job
from biolib.biolib_binary_format import ModuleInput
from biolib.cli_utils import BiolibValidationError
from biolib.validators.validate_zip_file import validate_zip_file  # type: ignore
from biolib.biolib_docker_client import BiolibDockerClient
from biolib.biolib_logging import logger
from biolib.lfs import create_large_file_system, push_large_file_system


class IllegalArgumentError(ValueError):
    pass


def file_path(path):
    if path.startswith('/'):
        full_path = path
    else:
        full_path = os.path.normpath(os.path.join(os.getcwd(), path))
    if not os.path.exists(full_path):
        raise IllegalArgumentError(f'The path {full_path} does not exist')
    return full_path


def port_number(port):
    if not port.isdigit():
        raise IllegalArgumentError(f'Port number {port} is not a number. Ports can only be numbers')

    if not (0 < int(port) < 65000):  # pylint: disable=superfluous-parens
        raise IllegalArgumentError('Port can only be between 0 and 65000')

    return port


def main():
    # set more restrictive default log level for CLI
    logger.configure(default_log_level=logging.WARNING)

    def _get_stdin():
        stdin = None
        if not sys.stdin.isatty() and not utils.IS_DEV:
            stdin = sys.stdin.read()
        return stdin

    if len(sys.argv) > 2 and sys.argv[1] == 'run':
        try:
            app = BioLibApp(uri=sys.argv[2])
        except biolib_errors.BioLibError as error:
            print(f'An error occurred:\n {error.message}', file=sys.stderr)
            sys.exit(1)

        app_args = sys.argv[3:]
        result = app.cli(args=app_args, stdin=_get_stdin(), files=None)
        result.save_files('biolib_results')
        sys.stdout.buffer.write(result.stdout)
        sys.stderr.buffer.write(result.stderr)
        sys.exit(result.exitcode)

    elif len(sys.argv) > 2 and sys.argv[1] == 'run-dev':
        provided_app_path = sys.argv[2]
        app_path = cli_utils.validate_and_get_app_path(provided_app_path)
        logger.info(f'Running BioLib application in local directory {app_path}...')
        tmp_directory, source_files_zip_path = cli_utils.create_temporary_directory_with_source_files_zip(app_path)

        try:
            validate_zip_file(source_files_zip_path, app_path)
            yaml_data = cli_utils.get_yaml_data(app_path)
            cli_utils.validate_yaml_config(yaml_data, source_files_zip_path)
            app_version_mocked = cli_utils.get_mocked_app_version_from_yaml(yaml_data, source_files_zip_path)

            for module in app_version_mocked['modules']:
                try:
                    BiolibDockerClient.get_docker_client().images.get(module['image_uri'])
                except ImageNotFound:
                    raise Exception(
                        f"Could not find local docker image {module['image_uri']} specified in .biolib/config.yml"
                    ) from None

            app_args = sys.argv[3:]
            args = cli_utils.validate_and_get_args(app_args)
            files_dict, args = cli_utils.get_files_dict_and_file_args(files=None, args=args)

            module_input_serialized = ModuleInput().serialize(
                stdin=_get_stdin() or b'',
                files=files_dict,
                arguments=args,
            )

            module_output = run_job(
                module_input_serialized=module_input_serialized,
                job=Job(
                    app_version=app_version_mocked,
                    caller_job=None,
                    created_at='',
                    public_id=utils.RUN_DEV_JOB_ID,
                    remote_hosts_with_warning=[],
                    user_id=None,
                ),
            )

            cli_utils.write_output_files(module_output['files'])

            print(cli_utils.get_pretty_print_module_output_string(
                exitcode=module_output['exit_code'],
                stderr=module_output['stderr'],
                stdout=module_output['stdout'],
            ))

        except Exception as exception:
            # Exit on BiolibValidationError as we have already printed the validation errors
            if isinstance(exception, BiolibValidationError):
                logger.error('Validation check failed for config file at .biolib/config.yml')
                sys.exit(1)
            raise exception

        finally:
            try:
                tmp_directory.cleanup()
            except IOError:
                logger.error('Failed to clean up temporary source files zip directory')

    else:
        parser = argparse.ArgumentParser()
        parser.add_argument('--version', action='version', version=utils.BIOLIB_PACKAGE_VERSION)

        subparsers = parser.add_subparsers(help='command', dest='command')

        # Add subparser for run to help message makes sense
        # The actual code for running applications is above this
        _parser_run = subparsers.add_parser('run', help='Run an application on BioLib')

        # Add subparser for run to help message makes sense
        # The actual code for running local applications is above this
        _parser_run_dev = subparsers.add_parser('run-dev', help='Run an application from a local directory')

        # add subparser for push
        parser_push = subparsers.add_parser('push', help='Push an application to BioLib')
        parser_push.add_argument('app_uri')
        parser_push.add_argument('--path', default='.', required=False)

        # add subparser for run-compute-node
        parser_start = subparsers.add_parser('start', help='Start a compute node')
        parser_start.add_argument('--port', default='5000', required=False, type=port_number)
        parser_start.add_argument('--host', default='127.0.0.1', required=False)

        # add subparser for lfs
        parser_lfs = subparsers.add_parser('lfs', help='Work with large file systems')
        subparsers_lfs = parser_lfs.add_subparsers(help='lfs command', dest='lfs_command')

        parser_lfs_create = subparsers_lfs.add_parser('create', help='Create a large filesystem')
        parser_lfs_create.add_argument('lfs_uri', type=str)

        parser_lfs_push = subparsers_lfs.add_parser('push', help='Push data to a large filesystem')
        parser_lfs_push.add_argument('lfs_uri', type=str)
        parser_lfs_push.add_argument('--path', type=str)

        args = parser.parse_args()

        logger.configure(default_log_level=logging.INFO)
        if args.command == 'push':
            push_application(args.app_uri, args.path)

        elif args.command == 'start':
            if platform.system() == 'Windows':
                raise Exception('Starting a compute node is currently not supported on Windows')

            from biolib.compute_node.webserver import webserver  # pylint: disable=import-outside-toplevel
            webserver.start_webserver(host=args.host, port=args.port)

        elif args.command == 'lfs':
            if args.lfs_command == 'create':
                create_large_file_system(
                    lfs_uri=args.lfs_uri,
                )
            elif args.lfs_command == 'push':
                push_large_file_system(
                    lfs_uri=args.lfs_uri,
                    input_dir=args.path
                )

        else:
            print('Unrecognized command, please run biolib --help to see available options.')
            sys.exit(1)


# allow this script to be called without poetry in dev e.g. by an IDE debugger
if utils.IS_DEV and __name__ == '__main__':
    main()



import base64

# necessary for making RSA import work TODO: figure out if this can be removed
from Crypto.IO import PEM  # pylint: disable=redefined-builtin, unused-import

from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_binary_format import RsaEncryptedAesPackage, AesEncryptedPackage
from biolib.biolib_logging import logger
from biolib.compute_node.job_worker.executors.types import RemoteExecuteOptions
from biolib.compute_node.job_worker.executors.remote.nitro_enclave_utils import NitroEnclaveUtils


class RemoteExecutor:

    @staticmethod
    def execute_job(options: RemoteExecuteOptions, module_input_serialized: bytes) -> bytes:
        job_id = options['job']['public_id']
        cloud_job = BiolibJobApi.create_cloud_job(module_name='main', job_id=job_id)
        logger.debug(f"Cloud: Job created with id {cloud_job['public_id']}")
        node_url = cloud_job['compute_node_info']['url']
        if 'attestation_document_base64' in cloud_job['compute_node_info']:
            attestation_document_bytes = base64.b64decode(cloud_job['compute_node_info']['attestation_document_base64'])
            expected_pcrs_and_aws_cert = BiolibJobApi.get_enclave_json(options['biolib_base_url'])

            rsa_public_key_der = NitroEnclaveUtils().attest_enclave_and_get_rsa_public_key(
                expected_pcrs_and_aws_cert,
                attestation_document_bytes,
            )
            serialized_data_to_send, aes_key_buffer = RsaEncryptedAesPackage().create(
                rsa_public_key_der,
                module_input_serialized,
            )
        else:
            serialized_data_to_send = module_input_serialized

        BiolibJobApi.start_cloud_job(job_id, serialized_data_to_send, node_url)
        BiolibJobApi.await_compute_node_status(
            compute_type='Cloud',
            job_id=job_id,
            node_url=node_url,
            retry_interval_seconds=1.5,
            retry_limit_minutes=30,
            status_to_await='Result Ready',
        )
        compute_result = BiolibJobApi.get_cloud_result(job_id, node_url)

        if 'attestation_document_base64' in cloud_job['compute_node_info']:
            serialized_module_output: bytes = AesEncryptedPackage(compute_result).decrypt(aes_key_buffer)
        else:
            serialized_module_output = compute_result
        return serialized_module_output



import asyncio
from asyncio import Future, Task
from pathlib import Path
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Generator, List, Optional, Set, Union

from biolib.pyppeteer.pyppeteer import helpers
from biolib.pyppeteer.pyppeteer.errors import BrowserError, NetworkError, PageError
from biolib.pyppeteer.pyppeteer.lifecycle_watcher import LifecycleWatcher
from biolib.pyppeteer.pyppeteer.models import JSFunctionArg, MouseButton
from biolib.pyppeteer.pyppeteer.timeout_settings import TimeoutSettings

if TYPE_CHECKING:
    from pyppeteer.jshandle import JSHandle, ElementHandle
    from pyppeteer.frame import Frame, FrameManager
    from pyppeteer.execution_context import ExecutionContext


async def readFileAsync(path, encoding):
    return Path(path).read_text(encoding=encoding)


class DOMWorld:
    def __init__(
        self, frameManager: 'FrameManager', frame: 'Frame', timeoutSettings: TimeoutSettings,
    ):
        self._frameManager = frameManager
        self._frame = frame
        self._timeoutSettings = timeoutSettings
        self.loop = self._frameManager._client.loop

        self._documentTask: Optional[Task['ElementHandle']] = None
        self._contextFuture: Optional[Future['ExecutionContext']] = None
        self._contextResolveCallback: Optional[Callable] = None
        self._setContext()

        self._waitTasks: Set[WaitTask] = set()
        self._detached = False

        # Aliases for query methods:
        self.J = self.querySelector
        self.Jx = self.xpath
        self.Jeval = self.querySelectorEval
        self.JJ = self.querySelectorAll
        self.JJeval = self.querySelectorAllEval

    @property
    def frame(self) -> 'Frame':
        return self._frame

    def _setContext(self, context: Optional['ExecutionContext'] = None) -> None:
        if context:
            self._contextResolveCallback(context)
            self._contextResolveCallback = None
            for waitTask in self._waitTasks:
                self.loop.create_task(waitTask.rerun())
        else:
            self._documentTask = None
            self._contextFuture = self.loop.create_future()
            self._contextResolveCallback = lambda _: self._contextFuture.set_result(_)

    def _hasContext(self) -> bool:
        return not self._contextResolveCallback

    def _detach(self) -> None:
        self._detached = True
        for task in self._waitTasks:
            task.terminate(BrowserError('waitForFunctions failed: frame got detached.'))

    @property
    async def executionContext(self) -> Optional['ExecutionContext']:
        if self._detached:
            raise BrowserError(
                'Execution Context is not available in detached '
                f'frame: {self._frame.url} (are you trying to evaluate?)'
            )
        return await self._contextFuture if self._contextFuture else None

    async def evaluateHandle(self, pageFunction: str, *args: JSFunctionArg) -> 'ElementHandle':
        context = await self.executionContext
        return await context.evaluateHandle(pageFunction=pageFunction, *args)

    async def evaluate(self, pageFunction: str, *args: JSFunctionArg):
        context = await self.executionContext
        return await context.evaluate(pageFunction, *args)

    async def querySelector(self, selector: str) -> Optional['ElementHandle']:
        document = await self._document
        return await document.querySelector(selector)

    @property
    def _document(self) -> Awaitable['ElementHandle']:
        if self._documentTask:
            return self._documentTask

        async def create_document_fut() -> 'ElementHandle':
            nonlocal self
            context = await self.executionContext
            document = await context.evaluateHandle('document')
            return document.asElement()

        self._documentTask = self.loop.create_task(create_document_fut())
        return self._documentTask

    async def xpath(self, expression: str) -> List['ElementHandle']:
        document = await self._document
        return await document.xpath(expression)

    async def querySelectorEval(self, selector: str, pageFunction: str, *args: JSFunctionArg) -> Any:
        document = await self._document
        return await document.querySelectorEval(selector, pageFunction, *args)

    async def querySelectorAll(self, selector: str) -> List['ElementHandle']:
        """Get all elements which matches `selector`.

        Details see :meth:`pyppeteer.page.Page.querySelectorAll`.
        """
        document = await self._document
        value = await document.querySelectorAll(selector)
        return value

    async def querySelectorAllEval(self, selector: str, pageFunction: str, *args: JSFunctionArg) -> Optional[Dict]:
        """Execute function on all elements which matches selector.

        Details see :meth:`pyppeteer.page.Page.querySelectorAllEval`.
        """
        document = await self._document
        value = await document.JJeval(selector, pageFunction, *args)
        return value

    @property
    async def content(self) -> str:
        return await self.evaluate(
            """
            () => {
              let retVal = '';
              if (document.doctype)
                retVal = new XMLSerializer().serializeToString(document.doctype);
              if (document.documentElement)
                retVal += document.documentElement.outerHTML;
              return retVal;
            }
            """
        )

    async def setContent(self, html: str, waitUntil=None, timeout=None) -> None:
        if timeout is None:
            timeout = self._timeoutSettings.navigationTimeout
        if waitUntil is None:
            waitUntil = ['load']
        await self.evaluate(
            """
            html => {
              document.open();
              document.write(html);
              document.close();
            }
            """,
            html,
        )
        watcher = LifecycleWatcher(
            frameManager=self._frameManager, frame=self._frame, waitUntil=waitUntil, timeout=timeout
        )
        error = await helpers.future_race(watcher.timeoutOrTerminationFuture, watcher.lifecycleFuture)
        watcher.dispose()
        if error:
            raise error

    async def addScriptTag(
        self,
        url: Optional[str] = None,
        path: Optional[Union[str, Path]] = None,
        content: Optional[str] = None,
        type_: str = '',
    ) -> Union['ElementHandle', 'JSHandle']:
        addScriptUrl = """
        async function addScriptUrl(url, type) {
          const script = document.createElement('script');
          script.src = url;
          if (type)
            script.type = type;
          const promise = new Promise((res, rej) => {
            script.onload = res;
            script.onerror = rej;
          });
          document.head.appendChild(script);
          await promise;
          return script;
        }
        """
        addScriptContent = """
        function addScriptContent(content, type = 'text/javascript') {
          const script = document.createElement('script');
          script.type = type;
          script.text = content;
          let error = null;
          script.onerror = e => error = e;
          document.head.appendChild(script);
          if (error)
            throw error;
          return script;
        }
        """
        context = await self.executionContext
        if url:
            try:
                return await context.evaluateHandle(addScriptUrl, url, type_)
            except Exception as e:
                raise BrowserError(f'Loading script from {url} failed: {e}')
        if path:
            path = Path(path)
            if not path.exists():
                raise FileNotFoundError(f'The specified file, {path.name}, does not exist')
            if not path.is_file():
                raise ValueError(f'The specified path, {path.name}, is not a file')
            contents = await readFileAsync(path, 'utf8')
            contents += '//# sourceURL' + path.name
            f = context.evaluateHandle(addScriptContent, contents, type_)
            return (await f).asElement()
        if content:
            f = context.evaluateHandle(addScriptContent, content, type_)
            return (await f).asElement()
        raise BrowserError('provide a url, path or content argument')

    async def addStyleTag(
        self, url: Optional[str] = None, path: Optional[Union[Path, str]] = None, content: Optional[str] = None
    ) -> 'ElementHandle':
        addStyleUrl = """
        async function addStyleUrl(url) {
          const link = document.createElement('link');
          link.rel = 'stylesheet';
          link.href = url;
          const promise = new Promise((res, rej) => {
            link.onload = res;
            link.onerror = rej;
          });
          document.head.appendChild(link);
          await promise;
          return link;
        }
        """
        addStyleContent = """
        async function addStyleContent(content) {
          const style = document.createElement('style');
          style.type = 'text/css';
          style.appendChild(document.createTextNode(content));
          const promise = new Promise((res, rej) => {
            style.onload = res;
            style.onerror = rej;
          });
          document.head.appendChild(style);
          await promise;
          return style;
        }
        """
        context = await self.executionContext
        if url:
            try:
                return (await context.evaluateHandle(addStyleUrl, url)).asElement()
            except Exception as e:
                raise BrowserError(f'Loading style from {url} failed')

        if path:
            path = Path(path)
            contents = await readFileAsync(path, 'utf8')
            contents += f'/*# sourceURL={path.name}*/'
            return (await context.evaluateHandle(addStyleContent, contents)).asElement()

        if content:
            return (await context.evaluateHandle(addStyleContent, content)).asElement()
        raise BrowserError('provide an object with url, path or content property')

    async def _select_handle(self, selector: str) -> 'ElementHandle':
        handle = await self.querySelector(selector)
        if not handle:
            raise BrowserError(f'No node found for selector: {selector}')
        return handle

    async def click(self, selector: str, button: MouseButton = 'left', clickCount: int = 1, delay: float = 0) -> None:
        """
        :param selector: CSS selector for element
        :param button:
        :param delay:
        :param clickCount:
        :return:
        """
        handle = await self._select_handle(selector)
        await handle.click(button=button, clickCount=clickCount, delay=delay)
        await handle.dispose()

    async def focus(self, selector: str) -> None:
        handle = await self._select_handle(selector)
        await handle.focus()
        await handle.dispose()

    async def hover(self, selector: str) -> None:
        handle = await self._select_handle(selector)
        await handle.hover()
        await handle.dispose()

    async def select(self, selector: str, *values: str) -> List[str]:
        handle = await self._select_handle(selector)
        result = await handle.select(*values)
        await handle.dispose()
        return result

    async def tap(self, selector: str) -> None:
        handle = await self._select_handle(selector)
        await handle.tap()
        await handle.dispose()

    async def type(self, selector: str, text: str, delay: float = 0) -> None:
        handle = await self._select_handle(selector)
        await handle.type(text=text, delay=delay)
        await handle.dispose()

    async def waitForSelector(
        self, selector: str, visible: bool = False, hidden: bool = False, timeout: float = None
    ) -> Optional['ElementHandle']:
        return await self._waitForSelectorOrXpath(
            selector, isXPath=False, visible=visible, hidden=hidden, timeout=timeout
        )

    async def waitForXpath(
        self, xpath: str, visible: bool = False, hidden: bool = False, timeout: float = None
    ) -> Optional['ElementHandle']:
        return await self._waitForSelectorOrXpath(xpath, isXPath=True, visible=visible, hidden=hidden, timeout=timeout)

    async def waitForFunction(
        self, pageFunction: str, *args: JSFunctionArg, polling: str = 'raf', timeout: Optional[float] = None,
    ) -> 'JSHandle':
        if not timeout:
            timeout = self._timeoutSettings.timeout
        return await WaitTask(self, pageFunction, 'function', polling, timeout, self.loop, *args).promise

    @property
    async def title(self) -> str:
        return await self.evaluate('document.title')

    async def _waitForSelectorOrXpath(
        self,
        selectorOrXpath: str,
        isXPath: bool,
        visible: bool = False,
        hidden: bool = False,
        timeout: Optional[float] = None,
    ) -> Optional['ElementHandle']:
        if not timeout:
            timeout = self._timeoutSettings.timeout
        if visible or hidden:
            polling = 'raf'
        else:
            polling = 'mutation'
        title = f"{'XPath' if isXPath else 'selector'} {selectorOrXpath}{' to be hidden' if hidden else ''}"
        predicate = """
        function predicate(selectorOrXPath, isXPath, waitForVisible, waitForHidden) {
          const node = isXPath
            ? document.evaluate(selectorOrXPath, document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue
            : document.querySelector(selectorOrXPath);
          if (!node)
            return waitForHidden;
          if (!waitForVisible && !waitForHidden)
            return node;
          const element = /** @type {Element} */ (node.nodeType === Node.TEXT_NODE ? node.parentElement : node);

          const style = window.getComputedStyle(element);
          const isVisible = style && style.visibility !== 'hidden' && hasVisibleBoundingBox();
          const success = (waitForVisible === isVisible || waitForHidden === !isVisible);
          return success ? node : null;

          /**
           * @return {boolean}
           */
          function hasVisibleBoundingBox() {
            const rect = element.getBoundingClientRect();
            return !!(rect.top || rect.bottom || rect.width || rect.height);
          }
        }
        """
        waitTask = WaitTask(
            self, predicate, title, polling, timeout, self.loop, selectorOrXpath, isXPath, visible, hidden
        )
        handle = await waitTask.promise
        if not handle.asElement():
            await handle.dispose()
            return None
        return handle.asElement()


class WaitTask:
    """WaitTask class.

    Instance of this class is awaitable.
    """

    def __init__(
        self,
        domWorld: DOMWorld,
        predicateBody: str,
        title: str,
        polling: Union[str, int],
        timeout: float,
        loop: asyncio.AbstractEventLoop,
        *args: JSFunctionArg,
    ) -> None:
        if isinstance(polling, str):
            if polling not in ['raf', 'mutation']:
                raise ValueError(f'Unknown polling: {polling}')
        elif isinstance(polling, (int, float)):
            if polling <= 0:
                raise ValueError(f'Cannot poll with non-positive interval: {polling}')
        else:
            raise ValueError(f'Unknown polling option: {polling}')

        self._domWorld = domWorld
        self._polling = polling
        self._timeout = timeout
        self.loop = loop
        if args or helpers.is_js_func(predicateBody):
            self._predicateBody = f'return ({predicateBody})(...args)'
        else:
            self._predicateBody = f'return {predicateBody}'
        self._args = args
        self._runCount = 0
        self._terminated = False
        self._timeoutError = False
        domWorld._waitTasks.add(self)

        self.promise = self.loop.create_future()

        async def timer(timeout: float) -> None:
            await asyncio.sleep(timeout / 1000)
            self._timeoutError = True
            self.terminate(TimeoutError(f'Waiting for {title} failed: timeout of {timeout}ms exceeded.'))

        if timeout:
            self._timeoutTimer = self.loop.create_task(timer(self._timeout))
        self._runningTask: Optional[Task] = self.loop.create_task(self.rerun())

    def __await__(self) -> Generator:
        """Make this class **awaitable**."""
        result = yield from self.promise
        if isinstance(result, Exception):
            raise result
        return result

    def terminate(self, error: Exception) -> None:
        """Terminate this task."""
        self._terminated = True
        if not self.promise.done():
            self.promise.set_exception(error)
        self._cleanup()

    async def rerun(self) -> None:  # noqa: C901
        """Start polling."""
        runCount = self._runCount = self._runCount + 1
        success: Optional['JSHandle'] = None
        error = None

        try:
            if await self._domWorld.executionContext is None:
                raise PageError(f'No execution context for {self._domWorld}')
            context = await self._domWorld.executionContext
            success = await context.evaluateHandle(
                waitForPredicatePageFunction, self._predicateBody, self._polling, self._timeout, *self._args,
            )
        except Exception as e:
            error = e

        if self.promise.done():
            return

        if self._terminated or runCount != self._runCount:
            if success:
                await success.dispose()
            return

        # Add try/except referring to puppeteer.
        try:
            if not error and success and (await self._domWorld.evaluate('s => !s', success)):
                await success.dispose()
                return
        except NetworkError:
            if success is not None:
                await success.dispose()
            return

        # page is navigated and context is destroyed.
        # Try again in the new execution context.
        if isinstance(error, NetworkError) and 'Execution context was destroyed' in error.args[0]:
            return

        # Try again in the new execution context.
        if isinstance(error, NetworkError) and 'Cannot find context with specified id' in error.args[0]:
            return

        if error:
            self.promise.set_exception(error)
        else:
            self.promise.set_result(success)

        self._cleanup()

    def _cleanup(self) -> None:
        if self._timeout and not self._timeoutError:
            self._timeoutTimer.cancel()
        self._domWorld._waitTasks.remove(self)
        self._runningTask = None


waitForPredicatePageFunction = """
async function waitForPredicatePageFunction(predicateBody, polling, timeout, ...args) {
  const predicate = new Function('...args', predicateBody);
  let timedOut = false;
  if (timeout)
    setTimeout(() => timedOut = true, timeout);
  if (polling === 'raf')
    return await pollRaf();
  if (polling === 'mutation')
    return await pollMutation();
  if (typeof polling === 'number')
    return await pollInterval(polling);

  /**
   * @return {!Promise<*>}
   */
  function pollMutation() {
    const success = predicate.apply(null, args);
    if (success)
      return Promise.resolve(success);

    let fulfill;
    const result = new Promise(x => fulfill = x);
    const observer = new MutationObserver(mutations => {
      if (timedOut) {
        observer.disconnect();
        fulfill();
      }
      const success = predicate.apply(null, args);
      if (success) {
        observer.disconnect();
        fulfill(success);
      }
    });
    observer.observe(document, {
      childList: true,
      subtree: true,
      attributes: true
    });
    return result;
  }

  /**
   * @return {!Promise<*>}
   */
  function pollRaf() {
    let fulfill;
    const result = new Promise(x => fulfill = x);
    onRaf();
    return result;

    function onRaf() {
      if (timedOut) {
        fulfill();
        return;
      }
      const success = predicate.apply(null, args);
      if (success)
        fulfill(success);
      else
        requestAnimationFrame(onRaf);
    }
  }

  /**
   * @param {number} pollInterval
   * @return {!Promise<*>}
   */
  function pollInterval(pollInterval) {
    let fulfill;
    const result = new Promise(x => fulfill = x);
    onTimeout();
    return result;

    function onTimeout() {
      if (timedOut) {
        fulfill();
        return;
      }
      const success = predicate.apply(null, args);
      if (success)
        fulfill(success);
      else
        setTimeout(onTimeout, pollInterval);
    }
  }
}
"""



import logging
import os
import sys
import random
import subprocess
import time
import pathlib

import requests.exceptions

from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_binary_format import ModuleOutput
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger


# TODO: type return value as ModuleOutput class
def run_job(job, module_input_serialized: bytes):
    host = '127.0.0.1'
    port = str(random.choice(range(5000, 65000)))
    node_url = f'http://{host}:{port}'
    job_id = job['public_id']
    logger.debug(f'Starting local compute node at {node_url}')
    start_cli_path = pathlib.Path(__file__).parent.parent.resolve() / 'start_cli.py'
    python_executable_path = sys.executable
    compute_node_process = subprocess.Popen(
        args=[python_executable_path, start_cli_path, 'start', '--host', host, '--port', port],
        env=dict(os.environ, BIOLIB_LOG=logging.getLevelName(logger.level)),
    )
    try:
        for retry in range(20):
            time.sleep(1)
            try:
                BiolibJobApi.save_compute_node_job(
                    job=job,
                    module_name='main',
                    access_token=BiolibApiClient.get().access_token,
                    node_url=node_url
                )
                break

            except requests.exceptions.ConnectionError:
                if retry == 19:
                    raise BioLibError('Could not connect to local compute node') from None
                logger.debug('Could not connect to local compute node retrying...')

        BiolibJobApi.start_cloud_job(job_id, module_input_serialized, node_url)
        BiolibJobApi.await_compute_node_status(
            retry_interval_seconds=1.5,
            retry_limit_minutes=30,
            status_to_await='Result Ready',
            compute_type='Compute Node',
            job_id=job_id,
            node_url=node_url,
        )

        result = BiolibJobApi.get_cloud_result(job_id, node_url)
        return ModuleOutput(result).deserialize()

    finally:
        compute_node_process.terminate()



import time
import requests

from biolib.biolib_api_client.auth import BearerAuth
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_errors import BioLibError
from biolib.compute_node.utils import SystemExceptionCodeMap
from biolib.biolib_logging import logger
from biolib.utils import BIOLIB_PACKAGE_VERSION


class RetryException(Exception):
    pass


class BiolibJobApi:

    @staticmethod
    def create(app_version_id, caller_job=None):
        data = {
                'app_version_id': app_version_id,
                'client_type': 'biolib-python',
                'client_version': BIOLIB_PACKAGE_VERSION,
        }
        if caller_job:
            data['caller_job'] = caller_job

        response = requests.post(
            f'{BiolibApiClient.get().base_url}/api/jobs/',
            auth=BearerAuth(BiolibApiClient.get().access_token),
            json=data,
        )

        # TODO: Error handling with response object
        if not response.ok:
            raise BioLibError(response.content)

        return response.json()

    @staticmethod
    def update_state(job_id, state):
        response = requests.patch(
            f'{BiolibApiClient.get().base_url}/api/jobs/{job_id}/',
            json={'state': state},
            auth=BearerAuth(BiolibApiClient.get().access_token)
        )

        # TODO: Error handling with response object
        if not response.ok:
            raise BioLibError(response.content)

        return response.json()

    @staticmethod
    def create_cloud_job(module_name, job_id):
        response = None
        for retry in range(4):
            try:
                response = requests.post(
                    f'{BiolibApiClient.get().base_url}/api/jobs/cloud/',
                    json={'module_name': module_name, 'job_id': job_id},
                    auth=BearerAuth(BiolibApiClient.get().access_token)
                )

                if response.status_code == 503:
                    raise RetryException(response.content)
                # Handle possible validation errors from backend
                elif not response.ok:
                    raise BioLibError(response.text)

                break

            except RetryException as retry_exception:  # pylint: disable=broad-except
                if retry > 3:
                    raise BioLibError('Reached retry limit for cloud job creation') from retry_exception
                time.sleep(1)

        if not response:
            raise BioLibError('Could not create new cloud job')

        cloud_job = response.json()
        if cloud_job.get('is_compute_node_ready', False):
            return cloud_job

        max_retry_attempts = 25
        retry_interval_seconds = 10

        for _ in range(max_retry_attempts):
            response = requests.get(
                f'{BiolibApiClient.get().base_url}/api/jobs/cloud/{cloud_job["public_id"]}/status/',
                auth=BearerAuth(BiolibApiClient.get().access_token)
            )
            cloud_job = response.json()
            if cloud_job.get('is_compute_node_ready', False):
                return cloud_job

            logger.info('Cloud: Reserved compute node not ready, retrying...')
            time.sleep(retry_interval_seconds)

        raise BioLibError('Cloud: The reserved compute node was not ready in time')

    @staticmethod
    def save_compute_node_job(job, module_name, access_token, node_url):
        response = requests.post(
            f'{node_url}/v1/job/',
            json={'module_name': module_name, 'job': job, 'access_token': access_token}
        )

        if not response.ok:
            raise BioLibError(response.content)

        return response.json

    @staticmethod
    def start_cloud_job(job_id, module_input_serialized, node_url):
        response = requests.post(
            f'{node_url}/v1/job/{job_id}/start/',
            data=module_input_serialized
        )

        if not response.ok:
            raise BioLibError(response.content)

    @staticmethod
    def await_compute_node_status(retry_interval_seconds, retry_limit_minutes, status_to_await,
                                  compute_type, job_id, node_url):
        status_max_retry_attempts = int(retry_limit_minutes * 60 / retry_interval_seconds)
        status_reached = False
        for _ in range(status_max_retry_attempts):
            response = requests.get(f'{node_url}/v1/job/{job_id}/status/')
            if not response.ok:
                raise Exception(response.content)

            status_json = response.json()

            for status_update in status_json['status_updates']:
                if 'log_message' in status_update:
                    logger.info(f'{compute_type}: {status_update["log_message"]}')

                if status_update.get('log_message') == status_to_await:
                    status_reached = True

            if 'error_code' in status_json:
                error_code = status_json['error_code']
                error_message = SystemExceptionCodeMap.get(error_code, f'Unknown error code {error_code}')

                raise BioLibError(f'{compute_type}: {error_message}')

            if status_reached:
                return

            time.sleep(retry_interval_seconds)

        raise BioLibError(f'{compute_type}: Failed to get results: Retry limit exceeded')

    @staticmethod
    def get_cloud_result(job_id, node_url):
        response = requests.get(
            f'{node_url}/v1/job/{job_id}/result/',
        )

        if not response.ok:
            raise Exception(response.content)

        return response.content

    @staticmethod
    def get_enclave_json(biolib_base_url):
        response = requests.get(
            f'{biolib_base_url}/info-files/biolib-enclave.json',
        )
        return response.json()



from biolib.typing_utils import TypedDict


class LargeFileSystemVersion(TypedDict):
    uri: str
    uuid: str


class LargeFileSystem(TypedDict):
    uri: str
    uuid: str


class LfsVersionPresignedUploadUrlResponse(TypedDict):
    presigned_upload_url: str


class LfsUploadPartMetadata(TypedDict):
    PartNumber: int
    ETag: str



import os

from biolib import typing_utils
from biolib.app import BioLibApp as _BioLibApp
from biolib.biolib_logging import logger as _logger
from biolib.biolib_api_client import BiolibApiClient as _BioLibApiClient

import biolib.app
import biolib.cli
import biolib.utils


# ------------------------------------ Function definitions for public Python API ------------------------------------

def call_cli() -> None:
    biolib.cli.main()


def load(uri: str) -> _BioLibApp:
    return _BioLibApp(uri)


def set_api_base_url(api_base_url: str) -> None:
    _BioLibApiClient.initialize(base_url=api_base_url)


def set_api_token(api_token: str) -> None:
    _BioLibApiClient.get().login(api_token)


def set_log_level(level: typing_utils.Union[str, int]) -> None:
    _logger.setLevel(level)


# -------------------------------------------------- Configuration ---------------------------------------------------
__version__ = biolib.utils.BIOLIB_PACKAGE_VERSION
_logger.configure(default_log_level='INFO')

set_api_base_url(os.getenv('BIOLIB_BASE_URL', default='https://biolib.com'))
set_api_token(os.getenv('BIOLIB_TOKEN', default=''))
